/* Copyright (c) 2012-2013 LevelUP contributors
 * See list at <https://github.com/rvagg/node-levelup#contributing>
 * MIT +no-false-attribs License <https://github.com/rvagg/node-levelup/blob/master/LICENSE>
 */

var Stream       = require('stream').Stream
  , bufferStream = require('simple-bufferstream')
  , inherits     = require('util').inherits
  , extend       = require('xtend')
  , errors       = require('./errors')
  , State        = require('./read-stream-state')
  , util         = require('./util')

  , defaultOptions = { keys: true, values: true }

  , makeKeyValueData = function (key, value) {
      return {
          key: util.decodeKey(key, this._options)
        , value: util.decodeValue(value, this._options)
      }
    }
  , makeKeyData = function (key) {
      return util.decodeKey(key, this._options)
    }
  , makeValueData = function (_, value) {
      return util.decodeValue(value, this._options)
    }
  , makeNoData = function () { return null }

function ReadStream (options, db, iteratorFactory) {
  Stream.call(this)

  this._state = State()

  this._dataEvent = 'data'
  this.readable = true
  this.writable = false

  // purely to keep `db` around until we're done so it's not GCed if the user doesn't keep a ref
  this._db = db

  options = this._options = extend(defaultOptions, options)

  this._keyEncoding   = options.keyEncoding   || options.encoding
  this._valueEncoding = options.valueEncoding || options.encoding

  if (typeof this._options.start != 'undefined')
    this._options.start = util.encodeKey(this._options.start, this._options)
  if (typeof this._options.end != 'undefined')
    this._options.end = util.encodeKey(this._options.end, this._options)
  if (typeof this._options.limit != 'number')
    this._options.limit = -1

  this._options.keyAsBuffer   = util.isKeyAsBuffer(this._options)

  this._options.valueAsBuffer = util.isValueAsBuffer(this._options)

  this._makeData = this._options.keys && this._options.values
    ? makeKeyValueData : this._options.keys
      ? makeKeyData : this._options.values
        ? makeValueData : makeNoData

  var self = this
  var ready = function () {
    if (!self._state.canEmitData())
      return

    self._state.ready()
    self._iterator = iteratorFactory(self._options)
    self._read()
  }

  if (db.isOpen())
    ready()
  else
    db.once('ready', ready)
}

inherits(ReadStream, Stream)

ReadStream.prototype.destroy = function () {
  this._state.destroy()
  if (this._state.canCleanup())
    this._cleanup()
}

ReadStream.prototype.pause = function () {
  if (this._state.canPause()) {
    this._state.pause()
    this.emit('pause')
  }
}

ReadStream.prototype.resume = function () {
  if (this._state.canResume()) {
    this.emit('resume')
    this._state.resume()
    this._read()
  }
}

ReadStream.prototype.pipe = function (dest) {
  if (typeof dest.add == 'function' && this._options.type == 'fstream') {
    this._dataEvent = 'entry'
    var self = this
    this.on('entry', function (data) {
      var entry = bufferStream(new Buffer(data.value))
      entry.path = data.key.toString()
      entry.type = 'File'
      entry.props = {
          type: 'File'
        , path: data.key.toString()
      }
      entry.pause()
      if (dest.add(entry) === false)
        self.pause()
    })
  }
  return Stream.prototype.pipe.apply(this, arguments)
}

ReadStream.prototype._read = function () {
  var self = this
  if (this._state.canRead()) {
    this._state.read()
    this._iterator.next(function(err, key, value) {
      self._onData(err, key, value)
    })
  }
}

ReadStream.prototype._onData = function (err, key, value) {
  this._state.endRead()
  if (err || (key === undefined && value === undefined) || !this._state.canEmitData())
    return this._cleanup(err)
  this._read() // queue another read even tho we may not need it
  try {
    value = this._makeData(key, value)
  } catch (e) {
    return this.emit('error', new errors.EncodingError(e))
  }
  this.emit(this._dataEvent, value)
}

ReadStream.prototype._cleanup = function (err) {
  if (err)
    this.emit('error', err)

  if (!this._state.canEnd())
    return

  this._state.end()
  this.readable = false

  this.emit('end')

  if (this._iterator) {
    var self = this
    this._iterator.end(function () {
      self._iterator = null
      self.emit('close')
    })
  } else
    this.emit('close')
}

ReadStream.prototype.toString = function () {
  return 'LevelUP.ReadStream'
}

module.exports.create = function (options, db, iteratorFactory) {
  return new ReadStream(options, db, iteratorFactory)
}
